/*
 * USBasp - USB in-circuit programmer for Atmel AVR controllers
 *
 * Thomas Fischl <tfischl@gmx.de>
 *
 * License........: GNU GPL v2 (see Readme.txt)
 * Target.........: ATMega8 at 12 MHz
 * Creation Date..: 2005-02-20
 * Last change....: 2009-02-28
 *
 * PC2 SCK speed option.
 * GND  -> slow (8khz SCK),
 * open -> software set speed (default is 375kHz SCK)
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbasp.h"
#include "usbdrv.h"
#include "isp.h"
#include "clock.h"

#define CONCATx(a,b) a##b
#define CONCAT(a,b) CONCATx(a,b)

#define USB_CFG_IODDR CONCAT(DDR, USB_CFG_IOPORTNAME)

static uchar replyBuffer[8];

static uchar prog_state = PROG_STATE_IDLE;
static uchar prog_sck = USBASP_ISP_SCK_AUTO;

static uchar prog_address_newmode = 0;
static unsigned long prog_address;
static unsigned int prog_nbytes = 0;
static unsigned int prog_pagesize;
static uchar prog_blockflags;
static uchar prog_pagecounter;


uchar usbFunctionSetup(uchar data[8]) {

  usbRequest_t *rq = (usbRequest_t *)data;   // cast to structured data for parsing
  usbMsgPtr = replyBuffer;

  switch (rq->bRequest) {
     case USBASP_FUNC_CONNECT:
         ispSetSCKOption(prog_sck);
         /* set compatibility mode of address delivering */
         prog_address_newmode = 0;
         ledRedOn();
         ispConnect();
         return 0;

     case USBASP_FUNC_DISCONNECT:
         ispDisconnect();
         ledRedOff();
         return 0;

     case USBASP_FUNC_TRANSMIT:
         replyBuffer[0] = ispTransmit(rq->wValue.bytes[0]);
         replyBuffer[1] = ispTransmit(rq->wValue.bytes[1]);
         replyBuffer[2] = ispTransmit(rq->wIndex.bytes[0]);
         replyBuffer[3] = ispTransmit(rq->wIndex.bytes[1]);
         return 4;

     case USBASP_FUNC_READFLASH:
         if (!prog_address_newmode)
             prog_address = rq->wValue.word;
         prog_nbytes = rq->wLength.word;
         prog_state = PROG_STATE_READFLASH;
         return 0xff; /* multiple in */
                 
     case USBASP_FUNC_READEEPROM:
         if (!prog_address_newmode)
             prog_address = rq->wValue.word;
         prog_nbytes = rq->wLength.word;
         prog_state = PROG_STATE_READEEPROM;
         return 0xff; /* multiple in */

     case USBASP_FUNC_ENABLEPROG:
         replyBuffer[0] = ispEnterProgrammingMode();
         return 1;

     case USBASP_FUNC_WRITEFLASH:
         if (!prog_address_newmode)
             prog_address = rq->wValue.word;
         prog_pagesize = rq->wIndex.bytes[0];
         prog_blockflags = rq->wIndex.bytes[1] & 0x0F;
         prog_pagesize += (((unsigned int) rq->wIndex.bytes[1] & 0xF0) << 4);
         if (prog_blockflags & PROG_BLOCKFLAG_FIRST) {
             prog_pagecounter = prog_pagesize;
         }
         prog_nbytes = rq->wLength.word;
         prog_state = PROG_STATE_WRITEFLASH;
         return 0xff; /* multiple out */

     case USBASP_FUNC_WRITEEEPROM:
         if (!prog_address_newmode)
             prog_address = rq->wValue.word;
         prog_pagesize = 0;
         prog_blockflags = 0;
         prog_nbytes = rq->wLength.word;
         prog_state = PROG_STATE_WRITEEEPROM;
         return 0xff; /* multiple out */

     case USBASP_FUNC_SETLONGADDRESS:
         /* set new mode of address delivering (ignore address delivered in commands) */
         prog_address_newmode = 1;
         /* set new address */
         prog_address = *((unsigned long *) &data[2]);
         return 0;

     case USBASP_FUNC_SETISPSCK:
         /* set sck option */
         prog_sck = rq->wValue.bytes[0];
         replyBuffer[0] = 0;
         return 1;
         
     case USBASP_FUNC_EPPULSE:
         isp_ep_pulse = rq->wValue.bytes[0];
         replyBuffer[0] = 0;
         return 1;
    
     default:
         return 0;
    }
}

uchar usbFunctionRead(uchar *data, uchar len) {
    uchar i;
    
    /* check if programmer is in correct read state */
    if ((prog_state != PROG_STATE_READFLASH) && (prog_state
        != PROG_STATE_READEEPROM)) {
        return 0xff;
    }
    
    /* fill packet */
    if (prog_state == PROG_STATE_READFLASH) {
        for (i = 0; i < len; i++) {
            data[i] = ispReadFlash(prog_address);            
            prog_address++;        
        }
    } else {
        for (i = 0; i < len; i++) {
            data[i] = ispReadEEPROM(prog_address);            
            prog_address++;        
        }        
    }
    
    /* last packet? */
    if (len < 8) {
        prog_state = PROG_STATE_IDLE;        
    }
    
    return len;
}

uchar usbFunctionWrite(uchar *data, uchar len) {
    uchar retVal = 0;
    uchar i;
    
    /* check if programmer is in correct write state */
    if ((prog_state != PROG_STATE_WRITEFLASH) && (prog_state
        != PROG_STATE_WRITEEEPROM)) {
        return 0xff;            
    }

    if ((prog_state == PROG_STATE_WRITEFLASH) && (prog_pagesize != 0)) {
        /* nur noch ispTransmit verwendet */
        for (i = 0; i < len; i++) {
            /* ispWriteFlash(prog_address, data[i], 0); */
            ispTransmit(0x40 | ((prog_address & 1) << 3));
            ispTransmit(0);
            ispTransmit(prog_address >> 1);
            ispTransmit(data[i]);
            /* ispWriteFlash */
            prog_pagecounter--;
            if (prog_pagecounter == 0) {                    
                /* ispFlushPage(prog_address, data[i]);  */
                ispTransmit(0x4C);
                ispTransmit(prog_address >> 9);
                ispTransmit(prog_address >> 1);
                ispTransmit(0);
                uint8_t pollvalue = data[i];
                if (pollvalue == 0xFF) {
                    clockWait(15);                    
                } else {
                    /* polling flash */
                    uchar retries = 30;
                    uint8_t starttime = TIMERVALUE;
                    while (retries != 0) {
                        ispTransmit(0x20 | ((prog_address & 1) << 3));
                        ispTransmit(prog_address >> 9);
                        ispTransmit(prog_address >> 1);
                        if (ispTransmit(0) != 0xFF) {
                        //if (ispReadFlash(prog_address) != 0xFF) {
                            break;                            
                        }
                        if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us) {
                            starttime = TIMERVALUE;
                            retries--;                            
                        }                        
                    }
                }
                /* ispFlushPage(prog_address, data[i]);  */
                prog_pagecounter = prog_pagesize;                    
            }                          
            
            prog_nbytes--;
            
            if (prog_nbytes == 0) {
                prog_state = PROG_STATE_IDLE;
                if ((prog_blockflags & PROG_BLOCKFLAG_LAST) && (prog_pagecounter
                    != prog_pagesize)) {
                    /* last block and page flush pending, so flush it now */
                    /* ispFlushPage(prog_address, data[i]);      */
                    ispTransmit(0x4C);
                    ispTransmit(prog_address >> 9);
                    ispTransmit(prog_address >> 1);
                    ispTransmit(0);
                    uint8_t pollvalue = data[i];
                    if (pollvalue == 0xFF) {
                        clockWait(15);                    
                    } else {
                        /* polling flash */
                        uchar retries = 30;
                        uint8_t starttime = TIMERVALUE;
                        while (retries != 0) {
                            ispTransmit(0x20 | ((prog_address & 1) << 3));
                            ispTransmit(prog_address >> 9);
                            ispTransmit(prog_address >> 1);
                            if (ispTransmit(0) != 0xFF) {
                            //if (ispReadFlash(prog_address) != 0xFF) {
                                break;                            
                            }
                            if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us) {
                                starttime = TIMERVALUE;
                                retries--;                            
                            }                        
                        }
                    }
                    /* ispFlushPage(prog_address, data[i]);  */                
                }
                retVal = 1; // Need to return 1 when no more data is to be received            
            }
            prog_address++;        
        } 
    } else {
        for (i = 0; i < len; i++) {
            if (prog_state == PROG_STATE_WRITEFLASH) {
                /* Flash */
                /* not paged */
                ispWriteFlash(prog_address, data[i], 1);                           
            } else {
                /* EEPROM */
                ispWriteEEPROM(prog_address, data[i]);            
            }
            
            prog_nbytes--;
            
            if (prog_nbytes == 0) {
                prog_state = PROG_STATE_IDLE;
                if ((prog_blockflags & PROG_BLOCKFLAG_LAST) && (prog_pagecounter
                    != prog_pagesize)) {
                    /* last block and page flush pending, so flush it now */
                    ispFlushPage(prog_address, data[i]);                    
                }
                retVal = 1; // Need to return 1 when no more data is to be received            
            }
            prog_address++;        
        }
    }
    return retVal;
}

/**************************************************************************
 * Mit externem 12 MHz Quarz ergibt sich eine Systemfrequenz
 * F_CPU von 12 MHz mit dem Vorteiler 1:1 (notwendig fuer USB-decoding
 *************************************************************************/
#ifdef __AVR_ATmega88__
void set_clk_prescaler(void) __attribute__ ((naked))
    __attribute__ ((section(".init5")));
void set_clk_prescaler(void)
{
    CLKPR = (1 << CLKPCE);
    CLKPR = 0;
}
#endif

int main(void) {
    
    uchar i, j;
    
    wdt_disable();
    
   
    /* no pullups on USB and ISP pins */
    PORTD = 0;
    PORTB = 0;
    
    /* all outputs except PD2 = INT0 */
    DDRD = ~(1 << 2);
    
    DDRC = (1 << PC1) | (1 << PC0);
    PORTC = (1 << PC2) | (1 << PC1) | (1 << PC0);
    
    //ledGreenOn();
    //return 0;
    
#ifdef USB_CFG_PULLUP_IOPORTNAME    
    usbDeviceDisconnect();   /* +ub -> 2k2 like on powerup */
    j = 0;
    /* USB Reset by device only required on Watchdog Reset */
    while (--j) {
        _delay_ms(2.0);            
    }
    usbDeviceConnect();
#else
    /* output SE0 for USB reset */
    USB_CFG_IODDR |= (1 << USB_CFG_DMINUS_BIT) | (1 << USB_CFG_DPLUS_BIT);
    j = 0;
    /* USB Reset by device only required on Watchdog Reset */
    while (--j) {
        _delay_ms(2.0);            
    }
    /* all USB and ISP pins inputs */
    USB_CFG_IODDR &= ~((1 << USB_CFG_DMINUS_BIT) | (1 << USB_CFG_DPLUS_BIT));
#endif
    
    /* init timer */
    clockInit();
    
#ifdef USBASP_HAS_CLOCK_GEN
#ifndef __AVR_ATmega8__
    /* 1 MHz generator megax8 */
    clockGenInit(0x05);
#endif
#endif
    
    /* main event loop */
    usbInit();
    wdt_enable(WDTO_1S);
    sei();
    ledGreenOn();
    for (;;) {
        wdt_reset();
        usbPoll();        
    }
    return 0;
}

